#ifndef MYX_REDIS_PUBSUB_INL_HPP_
#define MYX_REDIS_PUBSUB_INL_HPP_

#pragma once

#ifndef MYXLIB_HEADER_ONLY
#include <myx/redis/pubsub.hpp>
#endif

#include <myx/redis/containers.hpp>

#include <QTimer>
#include <QThread>
#include <QDateTime>
#include <QTcpSocket>
#include <QCoreApplication>

namespace myx {

namespace redis {

const QString PubSub::KeySpacePrefix { "__key%1__:" };

/*!
 * \brief Конструктор класса
 * \param parent Родительский объект для _socket
 */
PubSub::PubSub( QObject* parent ) :
	Base{ parent }
{
	_socket = new QTcpSocket { this };
	QObject::connect( _socket, &QAbstractSocket::stateChanged,
	                  this, &PubSub::onSocketStateChanged );
	connect( _socket, &QIODevice::readyRead, this, &PubSub::onReadyRead );
	connect( _socket, &QIODevice::bytesWritten, this, &PubSub::onBytesWritten );
	QObject::connect( _socket, static_cast< void ( QAbstractSocket::* )( QAbstractSocket::SocketError ) >
	                  ( &QAbstractSocket::error ), [this]() {
		qDebug() << QDateTime::currentDateTimeUtc().toString( "hh:mm:ss:zzz" )
		         << QByteArray( "redis pubsub connection error" ) << _socket->errorString();
	} );
	QObject::connect( _socket, &QIODevice::readyRead, this, &PubSub::read );
}


/*!
 * \brief Инициирует работу с БД
 */
void PubSub::start()
{
	emit disconnected();
	emit connectionStateChanged( _connectionFlags );
	_socket->connectToHost( _host, _port );
}


/*!
 * \brief Завершает работу с БД
 */
void PubSub::stop()
{
	if ( _socket->state() == QAbstractSocket::ConnectedState )
	{
		_socket->disconnectFromHost();
		while ( _socket->state() != QAbstractSocket::UnconnectedState ) QCoreApplication::processEvents();
	}
}


/*!
 * \brief Создает подписку на канал сообщений
 * \param channel Rанал подписки
 * \param subscriber Объект-подписчик
 * \param method Имя метода, вызываемого при изменении получении серверного сообщения
 */
void PubSub::subscribe( QString channel, QObject* subscriber, const char* method )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	auto iter = _subscribers.find( channel );
	if ( iter == _subscribers.end() )
	{
		if ( _socket->state() == QAbstractSocket::ConnectedState )
		{
			_socket->write( array( { "SUBSCIBE", channel } ) );
		}
		_subscribers[channel].push_back( { subscriber, method } );
	}
	else
	{
		auto value = iter.value();
		auto it    = std::find_if( value.begin(), value.end(),
		                           [subscriber, method]( const std::pair< QObject*, const char* >& p ) {
			return( p.first == subscriber && strcmp( p.second, method ) == 0 );
		} );
		if ( it == iter.value().end() )
		{
			iter.value().push_back( { subscriber, method } );
		}
	}
} // PubSub::subscribe


/*!
 * \brief Создает подписку на канал изменения данных
 * \param channel Rанал подписки
 * \param subscriber Объект-подписчик
 * \param method Имя метода, вызываемого при изменении получении серверного сообщения
 * \param database Номер БД
 */
void PubSub::subscribe( QString channel, QObject* subscriber, const char* method, int database )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	auto iter = _subscribers.find( channel );
	if ( iter == _subscribers.end() )
	{
		channel.prepend( KeySpacePrefix.arg( "space@" + QString::number( database ) ).toLocal8Bit() );
		if ( _socket->state() == QAbstractSocket::ConnectedState )
		{
			_socket->write( array( { "SUBSCRIBE", channel } ) );
		}
		_subscribers[channel].push_back( { subscriber, method } );
	}
	else
	{
		auto value = iter.value();
		auto it    = std::find_if( value.begin(), value.end(),
		                           [subscriber, method]( const std::pair< QObject*, const char* >& p ) {
			return( p.first == subscriber && strcmp( p.second, method ) == 0 );
		} );
		if ( it == iter.value().end() )
		{
			iter.value().push_back( { subscriber, method } );
		}
	}
} // PubSub::subscribe


/*!
 * \brief Создает подписку на шаблон канала сообщений
 * \param channel Шаблон канала подписки
 * \param subscriber Объект-подписчик
 * \param method Имя метода, вызываемого при изменении получении серверного сообщения
 */
void PubSub::psubscribe( QString channel, QObject* subscriber, const char* method )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	auto iter = _psubscribers.find( channel );
	if ( iter == _psubscribers.end() )
	{
		if ( _socket->state() == QAbstractSocket::ConnectedState )
		{
			_socket->write( array( { "PSUBSCRIBE", channel } ) );
		}
		_psubscribers[channel].push_back( { subscriber, method } );
	}
	else
	{
		auto value = iter.value();
		auto it    = std::find_if( value.begin(), value.end(),
		                           [subscriber, method]( const std::pair< QObject*, const char* >& p ) {
			return( p.first == subscriber && strcmp( p.second, method ) == 0 );
		} );
		if ( it == iter.value().end() )
		{
			iter.value().push_back( { subscriber, method } );
		}
	}
} // PubSub::psubscribe


/*!
 * \brief Создает подписку на шаблон канала изменения данных
 * \param channel Шаблон канала подписки
 * \param subscriber Объект-подписчик
 * \param method Имя метода, вызываемого при изменении получении серверного сообщения
 * \param database Номер БД (-1 для всех)
 */
void PubSub::psubscribe( QString channel, QObject* subscriber, const char* method, int database )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	auto iter = _psubscribers.find( channel );
	if ( iter == _psubscribers.end() )
	{
		if ( database == -1 )
		{
			channel.prepend( KeySpacePrefix.arg( "*" ).toLocal8Bit() );
		}
		else
		{
			channel.prepend( KeySpacePrefix.arg( "space@" + QString::number( database ) ).toLocal8Bit() );
		}
		if ( _socket->state() == QAbstractSocket::ConnectedState )
		{
			_socket->write( array( { "PSUBSCRIBE", channel } ) );
		}
		_psubscribers[channel].push_back( { subscriber, method } );
	}
	else
	{
		auto value = iter.value();
		auto it    = std::find_if( value.begin(), value.end(),
		                           [subscriber, method]( const std::pair< QObject*, const char* >& p ) {
			return( p.first == subscriber && strcmp( p.second, method ) == 0 );
		} );
		if ( it == iter.value().end() )
		{
			iter.value().push_back( { subscriber, method } );
		}
	}
} // PubSub::psubscribe


/*!
 * \brief Отписывает получателя от канала
 * \param channel Имя канала
 * \param subscriber Подписчик на канал
 * \param method Метод обработки данных
 * \param keyEvents Признак подписки на изменение данных
 */
void PubSub::unsubscribe( QString channel, QObject* subscriber, const char* method )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	auto iter = _subscribers.find( channel );
	if ( iter != _subscribers.end() )
	{
		auto ctx = std::find_if( iter->begin(), iter->end(),
		                         [subscriber, method]( const std::pair< QObject*, const char* >& value ) {
			return( subscriber == value.first && strcmp( method, value.second ) == 0 );
		} );
		if ( ctx != iter->end() )
		{
			iter->erase( ctx );
		}
		if ( iter->empty() )
		{
			if ( _socket->state() == QAbstractSocket::ConnectedState )
			{
				_socket->write( array( { "UNSUBSCRIBE", channel } ) );
			}
			_subscribers.remove( channel );
		}
	}
} // PubSub::unsubscribe


/*!
 * \brief Отписывает получателя от канала
 * \param channel Имя канала
 * \param subscriber Подписчик на канал
 * \param method Метод обработки данных
 * \param keyEvents Признак подписки на изменение данных
 */
void PubSub::unsubscribe( QString channel, QObject* subscriber, const char* method, int database )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	if ( database == -1 )
	{
		channel.prepend( KeySpacePrefix.arg( "*" ).toLocal8Bit() );
	}
	else
	{
		channel.prepend( KeySpacePrefix.arg( "space@" + QString::number( database ) ).toLocal8Bit() );
	}
	auto iter = _subscribers.find( channel );
	if ( iter != _subscribers.end() )
	{
		auto ctx = std::find_if( iter->begin(), iter->end(),
		                         [subscriber, method]( const std::pair< QObject*, const char* >& value ) {
			return( subscriber == value.first && strcmp( method, value.second ) == 0 );
		} );
		if ( ctx != iter->end() )
		{
			iter->erase( ctx );
		}
		if ( iter->empty() )
		{
			if ( _socket->state() == QAbstractSocket::ConnectedState )
			{
				_socket->write( array( { "UNSUBSCRIBE", channel } ) );
			}
			_subscribers.remove( channel );
		}
	}
} // PubSub::unsubscribe


/*!
 * \brief Отписывает получателя от шаблона канала
 * \param channel Шаблон канала
 * \param subscriber Подписчик на канал
 * \param method Метод обработки данных
 * \param keyEvents Признак подписки на изменение данных
 */
void PubSub::punsubscribe( QString channel, QObject* subscriber,
                           const char* method )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	auto iter = _psubscribers.find( channel );
	if ( iter != _psubscribers.end() )
	{
		auto ctx = std::find_if( iter->begin(), iter->end(),
		                         [subscriber, method]( const std::pair< QObject*, const char* >& value ) {
			return( subscriber == value.first && strcmp( method, value.second ) == 0 );
		} );
		if ( ctx != iter->end() )
		{
			iter->erase( ctx );
		}
		if ( iter->empty() )
		{
			if ( _socket->state() == QAbstractSocket::ConnectedState )
			{
				_socket->write( array( { "PUNSUBSCRIBE", channel } ) );
			}
			_psubscribers.remove( channel );
		}
	}
} // PubSub::punsubscribe


/*!
 * \brief Отписывает получателя от шаблона канала
 * \param channel Шаблон канала
 * \param subscriber Подписчик на канал
 * \param method Метод обработки данных
 * \param keyEvents Признак подписки на изменение данных
 */
void PubSub::punsubscribe( QString channel, QObject* subscriber,
                           const char* method, int database )
{
	if ( QThread::currentThread() != thread() )
	{
		qWarning() << "no redis connection for caller thread";
		return;
	}
	if ( database == -1 )
	{
		channel.prepend( KeySpacePrefix.arg( "*" ).toLocal8Bit() );
	}
	else
	{
		channel.prepend( KeySpacePrefix.arg( "space@" + QString::number( database ) ).toLocal8Bit() );
	}
	auto iter = _psubscribers.find( channel );
	if ( iter != _psubscribers.end() )
	{
		auto ctx = std::find_if( iter->begin(), iter->end(),
		                         [subscriber, method]( const std::pair< QObject*, const char* >& value ) {
			return( subscriber == value.first && strcmp( method, value.second ) == 0 );
		} );
		if ( ctx != iter->end() )
		{
			iter->erase( ctx );
		}
		if ( iter->empty() )
		{
			if ( _socket->state() == QAbstractSocket::ConnectedState )
			{
				_socket->write( array( { "PUNSUBSCRIBE", channel } ) );
			}
			_psubscribers.remove( channel );
		}
	}
} // PubSub::punsubscribe


/*!
 * \brief Разбирает поступившие от сервера уведомления и вызывает требуемые
 * метаметоды обработки
 */
void PubSub::read()
{
	int splitLength;
	while ( _socket->bytesAvailable() > 0 )
	{
		_buffer += _socket->readAll();
		while ( !_buffer.isEmpty() )
		{
			auto parts = split( _buffer, &splitLength );
			if ( parts.empty() )
			{
				break;
			}
			if ( ( parts[0] == "pmessage" ) && ( parts.size() == 4 ) )
			{
				auto iter = _psubscribers.find( parts[1] );
				if ( iter != _psubscribers.end() )
				{
					for ( const auto& i: iter.value() )
					{
						QMetaObject::invokeMethod( i.first, i.second,
						                           Q_ARG( QString, parts[2] ),
						                           Q_ARG( QByteArray, parts[3] ) );
					}
				}
			}
			else
			{
				if ( ( parts[0] == "message" ) && ( parts.size() == 3 ) )
				{
					auto iter = _subscribers.find( parts[1] );
					if ( iter != _subscribers.end() )
					{
						for ( const auto& i: iter.value() )
						{
							QMetaObject::invokeMethod( i.first, i.second,
							                           Q_ARG( QString, parts[1] ),
							                           Q_ARG( QByteArray, parts[2] ) );
						}
					}
				}
			}
			_buffer.remove( 0, splitLength );
		}
	}
} // PubSub::read


/*!
 * \brief Обработка изменения состояния сокета
 */
void PubSub::onSocketStateChanged()
{
	switch ( _socket->state() )
	{
	case QAbstractSocket::ConnectedState:
		for ( auto iter = _psubscribers.cbegin(); iter != _psubscribers.cend(); ++iter ) _socket->write( array( { "PSUBSCRIBE", iter.key() } ) );
		for ( auto iter = _subscribers.cbegin(); iter != _subscribers.cend(); ++iter ) _socket->write( array( { "SUBSCRIBE", iter.key() } ) );
		qInfo() << QDateTime::currentDateTimeUtc().toString( "hh:mm:ss:zzz" ) <<
		        QString( "redis at %1:%2 is connected" ).arg( _host ).arg( _port );
		emit connectionStateChanged( _connectionFlags = Connected );
		emit connected();
		break;
	case QAbstractSocket::UnconnectedState:
		_buffer.clear();
		QTimer::singleShot( Timeout, [this]() {
			_socket->connectToHost( _host, _port );
		} );
		qInfo() << QDateTime::currentDateTimeUtc().toString( "hh:mm:ss:zzz" ) <<
		        QString( "redis at %1:%2 is unconnected" ).arg( _host ).arg( _port );
		if ( _connectionFlags & Connected )
		{
			emit disconnected();
		}
		emit connectionStateChanged( _connectionFlags = Unconnected );
		break;
	default:
		break;
	} // switch
} // PubSub::onSocketStateChanged


/*!
 * \brief Возвращает признак установленного соединения
 * \return Признак установленного соединения
 */
bool PubSub::isConnected() const
{
	return ( _connectionFlags & Connected );
}


/*!
 * \brief Возвращает сокет
 * \return Сокет
 */
QTcpSocket* PubSub::socket() const
{
	return( _socket );
}

} // namespace redis

} // namespace myx


#endif // ifndef MYX_REDIS_PUBSUB_INL_HPP_
